/*
 * ReferenceDialog.java
 * Created on 2012/11/18
 *
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
package jp.co.ntt.oss.heapstats.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Point;
import java.awt.datatransfer.DataFlavor;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.SwingConstants;
import javax.swing.border.BevelBorder;

import jp.co.ntt.oss.heapstats.container.ObjectData;
import jp.co.ntt.oss.heapstats.container.SnapShot;
import jp.co.ntt.oss.heapstats.gui.snapshot.SnapShotReader;

import com.mxgraph.layout.hierarchical.mxHierarchicalLayout;
import com.mxgraph.model.mxCell;
import com.mxgraph.model.mxGeometry;
import com.mxgraph.model.mxICell;
import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.swing.util.mxGraphTransferable;

/**
 * extends {@link JDialog}<br>
 * implements {@link MouseListener}<br>
 * <br>
 * Draw the Object reference information.
 */
public class ReferenceDialog extends JDialog implements
                                            MouseListener {

    /** serialVersionUID. */
    private static final long serialVersionUID = -2484632032906316369L;

    /** Main Component. */
    private JPanel comp;

    /** Reference to the region for displaying the object. */
    private ReferenceGraph graph;

    /** Reference to the enclosing graph component. */
    private mxGraphComponent graphComponent;

    /** a snapshot Information. */
    private SnapShotReader snapShotReader;

    /** Search radio button to indicate the direction of the child. */
    private JRadioButton child;

    /** Search radio button to indicate the direction of the parent. */
    private JRadioButton parent;

    /** Radio button to set the sorting criteria. */
    private JRadioButton instance;

    /** Radio button to set the sorting criteria. */
    private JRadioButton heap;

    /** Button to enable condition. */
    private JButton exec;

    /**
     * Check box to set the terms of the number of objects to be deployed to a
     * single.
     */
    private JCheckBox rank;

    /** Reference direction holds whether or not the parent. */
    private boolean isParent;

    /** Sort condition holds whether the number of instances. */
    private boolean isInstance;

    /** Holds whether to limit the number of objects that you want to deploy. */
    private boolean isRank;

    /**
     * Create the 'Reference Dialog'.
     *
     * @param owner the owner Dialog from which the dialog is displayed or null
     *        if this dialog has no owner
     */
    public ReferenceDialog(final JFrame owner) {
        super(owner);
        setSize(GuiCommon.DEFAULT_DIALOG_WIDTH, GuiCommon.DEFAULT_DIALOG_HEIGHT);
        setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        setModal(false);
        setTitle(GuiCommon.getResources().getString(
                GuiCommon.TAB_TITLE_REFERENCE));

        comp = new JPanel(new BorderLayout());
        comp.setBackground(Color.WHITE);
        comp.setBorder(BorderFactory.createEmptyBorder(GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE, GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE));
        add(comp);
        setLocationByPlatform(true);
    }

    /**
     * Return the component.
     *
     * @return Return the component
     */
    public final JPanel getComp() {
        return comp;
    }

    /**
     * Sets the dialog component.
     *
     * @param value the Main Component
     */
    public final void setComp(final JPanel value) {
        comp = value;
        add(comp);
    }

    /**
     * This method does the drawing SnapShot Data.
     *
     * @param tag Sets the tag
     * @param reader a snapshot Information
     */
    public final void draw(final long tag, final SnapShotReader reader) {
        clear();

        snapShotReader = reader;
        JPanel panelInfo = drawInfo();

        if (panelInfo == null) {
            return;
        }

        graph = new ReferenceGraph();
        graph.setCellsEditable(false);

        graph.getModel().beginUpdate();

        try {
            SnapShot snap = snapShotReader.getClassEntry();
            ObjectData data = snap.getElements().get(tag);
            if (data != null) {
                long loader = data.getClassLoaderTag();
                if (loader < 0) {
                    data.setLoaderName("-");
                } else if (loader == 0) {
                    data.setLoaderName("<SystemClassLoader>");
                } else {
                    ObjectData loaderInfo = snap.getElements().get(loader);
                    if (loaderInfo != null) {
                        data.setLoaderName(loaderInfo.getName() + "(0x"
                                + Long.toHexString(data.getClassLoader()) + ")");
                    } else {
                        data.setLoaderName("-(0x"
                                + Long.toHexString(data.getClassLoader()) + ")");
                    }
                }

                ReferenceCell cell = new ReferenceCell(data, true, false);
                graph.addCell(cell);
            }
        } catch (IOException e) {
            GuiCommon.showMessage(this, GuiCommon.BASEE21,
                    JOptionPane.ERROR_MESSAGE);
            e.printStackTrace();
        }
        graph.getModel().endUpdate();

        graph.setEdgeStyle();

        graphComponent = new mxGraphComponent(graph);
        graphComponent.getGraphControl().addMouseListener(this);
        graphComponent.setToolTips(true);

        comp.add(panelInfo, BorderLayout.NORTH);
        comp.add(graphComponent, BorderLayout.CENTER);

        if (this.isVisible()) {
            comp.setVisible(true);
        }
    }

    /**
     * Clear the reference data.
     */
    public final void clear() {
        comp.setVisible(false);
        for (Component obj : comp.getComponents()) {
            comp.remove(obj);
        }
        isParent = true;
        isInstance = true;
        isRank = true;
    }

    @Override
    public final void mouseClicked(final MouseEvent arg0) {
        Object eventSource = arg0.getSource();
        if (eventSource.equals(graphComponent.getGraphControl())) {
            Object cell = graphComponent.getCellAt(arg0.getX(), arg0.getY());
            if (arg0.getButton() == MouseEvent.BUTTON1 && cell != null
                    && cell instanceof ReferenceCell) {
                drawMind((ReferenceCell) cell);
            }
        } else if (eventSource.equals(exec)) {
            isParent = parent.isSelected();
            isInstance = instance.isSelected();
            isRank = rank.isSelected();

            mxCell cell = (mxCell) graph.getDefaultParent();
            List<Object> removeCells = new ArrayList<Object>();

            for (int i = 0; i < cell.getChildCount(); i++) {
                mxCell removeCell = (mxCell) cell.getChildAt(i);
                if (removeCell.isEdge()
                        || !((ReferenceCell) removeCell).isRoot()) {
                    removeCells.add(removeCell);
                } else {
                    graph.removeSelectionCell(removeCell);
                    mxGeometry geometry = ((ReferenceCell) removeCell)
                            .getGeometry();
                    geometry.setX(0);
                    geometry.setY(0);
                    ((ReferenceCell) removeCell).setGeometry(geometry);
                }
            }
            graph.removeCells(removeCells.toArray());
            graph.refresh();
            graphComponent.refresh();
            graphComponent.getViewport().setViewPosition(new Point(0, 0));
        }
    }

    @Override
    public void mouseEntered(final MouseEvent arg0) {
    }

    @Override
    public void mouseExited(final MouseEvent arg0) {
    }

    @Override
    public void mousePressed(final MouseEvent arg0) {
    }

    @Override
    public void mouseReleased(final MouseEvent arg0) {
    }

    /**
     * Draw a condition setting panel.
     *
     * @return Return the panel
     */
    private JPanel drawInfo() {
        JPanel panelRange = new JPanel();
        panelRange.setLayout(new BoxLayout(panelRange, BoxLayout.X_AXIS));
        panelRange.setBackground(Color.WHITE);

        JLabel label = new JLabel();

        // Search Direction
        JPanel directionPanel = new JPanel();
        directionPanel.setBackground(Color.WHITE);
        directionPanel.setBorder(new BevelBorder(BevelBorder.LOWERED));
        parent = new JRadioButton(GuiCommon.getResources().getString(
                GuiCommon.DIRECTION_PARENT), true);
        child = new JRadioButton(GuiCommon.getResources().getString(
                GuiCommon.DIRECTION_CHILD));
        parent.setBackground(Color.WHITE);
        child.setBackground(Color.WHITE);
        ButtonGroup bg = new ButtonGroup();
        bg.add(parent);
        bg.add(child);
        directionPanel.add(new JLabel(GuiCommon.getResources().getString(
                GuiCommon.SEARCH_DIRECTION)));
        directionPanel.add(parent);
        directionPanel.add(child);

        // Sort
        JPanel sortPanel = new JPanel();
        sortPanel.setBackground(Color.WHITE);
        sortPanel.setBorder(new BevelBorder(BevelBorder.LOWERED));
        instance = new JRadioButton(GuiCommon.getResources().getString(
                GuiCommon.SORT_INSTANCE), true);
        heap = new JRadioButton(GuiCommon.getResources().getString(
                GuiCommon.SORT_SIZE), false);
        instance.setBackground(Color.WHITE);
        heap.setBackground(Color.WHITE);
        ButtonGroup rankGr = new ButtonGroup();
        rankGr.add(instance);
        rankGr.add(heap);
        sortPanel.add(new JLabel(GuiCommon.getResources().getString(
                GuiCommon.SORT_LABEL)));
        sortPanel.add(instance);
        sortPanel.add(heap);

        // Ranking
        rank = new JCheckBox(GuiCommon.getResources().getString(
                GuiCommon.CHECK_RANK), true);
        rank.setBackground(Color.WHITE);

        exec = new JButton("OK");
        exec.addMouseListener(this);

        try {
            SimpleDateFormat format = new SimpleDateFormat(
                    GuiCommon.DATE_FORMAT);
            label.setText(GuiCommon.getResources().getString(
                    GuiCommon.DATE_LABEL)
                    + "  :  "
                    + format.format(snapShotReader.getHeader()
                            .getSnapShotDate()));
        } catch (IOException e) {
            GuiCommon.showMessage(this, GuiCommon.BASEE21,
                    JOptionPane.ERROR_MESSAGE);
            e.printStackTrace();
            return null;
        }

        panelRange.add(label);
        panelRange.add(new JLabel(" "));
        panelRange.add(directionPanel);
        panelRange.add(new JLabel(" "));
        panelRange.add(sortPanel);
        panelRange.add(rank);
        panelRange.add(new JLabel(" "));
        panelRange.add(exec);

        return panelRange;
    }

    /**
     * Displayed on the graph to get a child or of a cell object you clicked,
     * the information of the parent.
     *
     * @param parentCell Cell object you clicked
     */
    private void drawMind(final ReferenceCell parentCell) {
        if (parentCell.getEdgeCount() > 1
                || (parentCell.isRoot() && parentCell.getEdgeCount() > 0)
                || parentCell.isEdge()) {
            return;
        }

        List<ObjectData> objectList = null;
        try {
            if (isParent) {
                SnapShot snapShot = snapShotReader.getToParents(parentCell
                        .getTag());
                objectList = snapShot.getToParents().get(parentCell.getTag());
            } else {
                SnapShot snapShot = snapShotReader.getToChildren(parentCell
                        .getTag());
                objectList = snapShot.getToChildren().get(parentCell.getTag());
            }
        } catch (FileNotFoundException e) {
            // There is also that the file does not exist, ignore this
            // exception.
            e.printStackTrace();
        } catch (Exception e) {
            GuiCommon.showMessage(this, GuiCommon.BASEE21,
                    JOptionPane.ERROR_MESSAGE);
            e.printStackTrace();
        }

        if (objectList == null) {
            return;
        }

        graph.getModel().beginUpdate();

        int j = 0;
        for (ObjectData entry : sort(objectList)) {
            ReferenceCell cell = null;
            mxCell tmp = (mxCell) parentCell.getParent();
            for (int i = 0; i < tmp.getChildCount(); i++) {
                mxICell target = tmp.getChildAt(i);
                if (target.isVertex()) {
                    if (((ReferenceCell) target).getTag() == entry.getTag()) {
                        cell = (ReferenceCell) target;
                    }
                }
            }

            ReferenceCell edge = new ReferenceCell(entry, false, true);

            if (cell == null) {
                cell = new ReferenceCell(entry, false, false);
                graph.addCell(cell);
            }
            long size = entry.getTotalSize() / GuiCommon.BINARY_PREFIX;
            if (size == 0) {
                edge.setValue(entry.getCount() + "\n< 1");
            } else {
                edge.setValue(entry.getCount() + "\n" + entry.getTotalSize()
                        / GuiCommon.BINARY_PREFIX);
            }

            graph.addEdge(edge, graph.getDefaultParent(), parentCell, cell, 1);

            if (isRank) {
                if (++j == GuiCommon.getRankLevel()) {
                    break;
                }
            }
        }

        if (mxGraphTransferable.dataFlavor == null) {
            try {
                mxGraphTransferable.dataFlavor = new DataFlavor(
                        DataFlavor.javaJVMLocalObjectMimeType
                                + "; class=com.mxgraph.swing.util.mxGraphTransferable",
                        null, new mxGraphTransferable(null, null).getClass()
                                .getClassLoader());
            } catch (ClassNotFoundException e1) {
                // do nothing
            }
        }

        mxHierarchicalLayout layout = new mxHierarchicalLayout(graph,
                SwingConstants.WEST);
        layout.execute(graph.getDefaultParent());
        graph.getModel().endUpdate();
    }

    /**
     * Returns a list of the object after the sort.
     *
     * @param source Object list before sorting
     * @return Returns a list of the object after the sort
     */
    private List<ObjectData> sort(final List<ObjectData> source) {
        List<ObjectData> viewList = new ArrayList<ObjectData>();

        for (ObjectData entry : source) {
            int i = 0;
            for (ObjectData view : viewList) {

                if (isInstance) {
                    if (view.getCount() < entry.getCount()) {
                        break;
                    }
                } else {
                    if (view.getTotalSize() < entry.getTotalSize()) {
                        break;
                    }
                }
                i++;
            }
            viewList.add(i, entry);
        }
        return viewList;
    }
}
